/*
 * Copyright (c) 1991 University of Southern California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the University of Southern California. The name of the University 
 * may not be used to endorse or promote products derived from this 
 * software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
*/

/*
 * $Log: generator.c,v $
 * Revision 1.4  1991/11/28  19:01:20  jamin
 * replace non-portable usleep with select.
 * Thanks to Jeff Mogul.
 *
 * Revision 1.3  1991/11/22  21:43:16  jamin
 * logfile named <hostname>.
 *
 * Revision 1.2  1991/11/22  07:26:37  jamin
 * added a -d option to specify duration.
 *
 * generator
 *
 *    this routine generates a random stream of UDP packets to the 
 * specified host, sending packets using an exponential distribution
 * with a specified mean.
 *    inputs: remote_host, mean_interarrival
 *
*/

#include <stdio.h>
#include <signal.h>
#include <strings.h>
#include <math.h>
#include <ctype.h>

#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>


double drand48();

/*
** locally defined routines
*/
void  usage();
void  parse_command_line();
void  signal_handler();
u_long  sleep_time();

/*
** global vars
*/
FILE *  log_file_fp;    /* log file filepointer */
u_long  send_total;    /* total number pkts sent */


main(argc, argv)
    int    argc;
    char **  argv;
{

    /*
    ** input params
    */
    char  remote_host[128];   /* host to send to */
    u_short  remote_port;    /* port to send to */
    u_long  mean_interarrival;    /* mean send interarrival in microseconds */
    long  duration;        /* run duration in seconds. */
    int    verbose;

    /*
    ** local vars
    */
    FILE *      date_fp;        /* file pointer to read date output from */
    char        date[256];      /* current date string */
    struct timeval sleep_interval;  /* time to sleep before next send */
    struct timeval starttime, now;   /* how long have we been running? */
    int    s;    /* send socket */
    struct sockaddr_in sin;  /* remote host internet address */
    struct hostent *hp;    /* host info for remote host */
    char *  addr_ptr;  /* pointer into host name/address string */
    u_long  host_num;  /* hold host address if entered as number */
    char  local_host[128];


    (void)strcpy(remote_host, "localhost");
    remote_port = (u_short)9;
    mean_interarrival = 1000000L;
    duration = 86700L;
    verbose = 0;

    /*
    ** parse the command line for any default overrides
    */
    (void)parse_command_line(argc, argv, remote_host,
      &remote_port, &mean_interarrival, &duration, &verbose);
    
    /*
    ** install the signal handler
    */
    (void)signal(SIGHUP, signal_handler);
    (void)signal(SIGINT, signal_handler);
    (void)signal(SIGQUIT, signal_handler);
    (void)signal(SIGTERM, signal_handler);

    /*
    ** set up socket to send on
    */
    if ((s = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
  perror("generator: socket");
  exit(1);
    }

    /*
    ** get the local hosts name
    */
    if (gethostname(local_host, sizeof(local_host)) == -1) {
        perror("generator: gethostname failed");
        exit(1);
    }

    /*
    ** open the log file
    */
    log_file_fp = fopen(local_host, "w");
    if (!log_file_fp) {
        (void)fprintf(stderr, "could not open log file %s\n", local_host);
        exit(1);
    }

    /*
    ** get the host entry structure using the looked up name
    */
    if ((hp = gethostbyname(local_host)) == (struct hostent *)NULL) {
        perror("generator: gethostbyname failed");
        exit(1);
    }

    /*
    ** bind local address
    */
    bzero((char *)&sin, sizeof(sin));
    sin.sin_family = hp->h_addrtype;
    sin.sin_port = htons(remote_port);
    bcopy(hp->h_addr, (char *)&sin.sin_addr, hp->h_length);
/*    if (bind(s, (struct sockaddr *)&sin, sizeof(sin)) < 0) {
  perror("generator: bind failed");
  exit(1);
    } */
    bzero((char *)&sin, sizeof(sin));

    /*
    ** get the remote host address
    */
    if ((int)(host_num = inet_addr(remote_host)) == -1) {
        if ((hp = gethostbyname(remote_host)) == (struct hostent *)NULL) {
            fprintf(stderr, "generator: %s unknown host\n", remote_host);
            exit(1);
        }
        bcopy(hp->h_addr, (char *)&sin.sin_addr, hp->h_length);
        sin.sin_family = hp->h_addrtype;
    }
    else {
  bcopy((char *)&host_num, (char *)&sin.sin_addr, sizeof(host_num));
  sin.sin_family = AF_INET;
    }
    sin.sin_port = htons(remote_port);

    /*
    ** print the start time
    */
    if ((date_fp = popen("date", "r")) == NULL) {
        perror("popen(date)");
    } else {
        (void)fgets(date, sizeof(date), date_fp);
        pclose(date_fp);
        fprintf(log_file_fp, "Start Time: %s\n", date);
    }
    fprintf(log_file_fp, "\tlocal host = %s\n", local_host);
    fprintf(log_file_fp, "\tremote host = %s\n", remote_host);
    fprintf(log_file_fp, "\tremote port = %d\n", remote_port);
    fprintf(log_file_fp, "\tmean interarrival = %u (usec)\n", mean_interarrival);
    fprintf(log_file_fp, "\tduration = %ld (sec)\n", duration);
    fprintf(log_file_fp, "\n");
    fflush(log_file_fp);


    /*
    ** send packets (forever)
    */
    send_total = 0;
    gettimeofday(&now, NULL);
    starttime = now;
    while (now.tv_sec - starttime.tv_sec < duration) {

      /*
      ** send next UDP packet
      */
      if (sendto(s, (char *)NULL, 0, 0, 
        (struct sockaddr *)&sin, sizeof(sin)) < 0) {
          perror("generate: sendto");
          fprintf(log_file_fp, "sendto() failed\n");
          exit(1);
      }
      send_total++;

      /*
      ** sleep until time to send next packet
      */
      sleep_interval.tv_usec = sleep_time(mean_interarrival);
      sleep_interval.tv_sec = sleep_interval.tv_usec / 1000000L;
      sleep_interval.tv_usec = sleep_interval.tv_usec % 1000000L;

      if (verbose) {
          printf("sleep time = %u:%u\n", sleep_interval.tv_sec, sleep_interval.tv_usec);
      }
       select(0, 0, 0, 0, &sleep_interval);
      /* usleep(sleep_interval); sleep_interval was an u_long */
      gettimeofday(&now, NULL);
    }
    signal_handler();
}

/*  */

/**********************************************************************
** usage
**
**    print out command usage info and exit
**********************************************************************/
void usage()
{
    fprintf(stderr, "generate {options}\n");
    fprintf(stderr, "  options:\n");
    fprintf(stderr, "\t[-h remote_host] : default = localhost\n");
    fprintf(stderr, "\t[-p remote_port] : default = 9\n");
    fprintf(stderr, "\t[-d duration (min)] : default = 1445\n");
    fprintf(stderr, "\t[-m mean_interarrival (sec)] : default = 1\n");
    fprintf(stderr, "\t[-v] : verbose\n");
    exit(1);
}

/*  */

/**********************************************************************
** parse_command_line
**
**    parse the command line and set any switches which are specified
**********************************************************************/
void parse_command_line(argc, argv, remote_host, remote_port, 
  mean_interarrival, duration, verbose)
    int         argc;
    char **     argv;
    char *  remote_host;
    u_short *  remote_port;
    u_long *  mean_interarrival;
    long *  duration;
    int *       verbose;
{
    argc--; argv++;
    while (argc) {

        if (!strcmp(*argv, "-h")) {
            argc--; argv++;
            (void)strcpy(remote_host, *argv);
        }
        else if (!strcmp(*argv, "-p")) {
            argc--; argv++;
            *remote_port = (u_short)atoi(*argv);
        }
        else if (!strcmp(*argv, "-m")) {
            argc--; argv++;
            *mean_interarrival = (u_long)atol(*argv) * 1000000L;
        }
        else if (!strcmp(*argv, "-d")) {
            argc--; argv++;
            *duration = (long)atol(*argv) * 60L;
        }
        else if (!strcmp(*argv, "-v")) {
            *verbose = 1;
        }
        else {
            usage();
        }
        argc--; argv++;
    }
    return;
}

/*  */

/**********************************************************************
** signal_handler
**
**    this routine is called when a hangup signal (SIGHUP) is sent to
** the process, it prints the stop time and logs the number of packets
** sent
**********************************************************************/
void signal_handler()
{

    FILE *      date_fp;
    char        date[256];

    /*
    ** print the stop time
    */
    if ((date_fp = popen("date", "r")) == NULL) {
        perror("popen(date)");
    } else {
        (void)fgets(date, sizeof(date), date_fp);
        pclose(date_fp);
        fprintf(log_file_fp, "Stop Time: %s\n", date);
    }
    fprintf(log_file_fp, "total UDP packets sent = %u\n", send_total);
    exit(0);
}

/*  */

/**********************************************************************
** sleep_time
**
**    this routine generates an exponentially distributed random number
** with the specified mean to be used as the length of time to sleep
** before sending the next packet
**********************************************************************/
u_long sleep_time(mean_interarrival)
    u_long   mean_interarrival;
{

    return ((u_long)((-(double)mean_interarrival) * log(drand48())));
}
